/* Allows two programs using the pseuto ttys to open and close their
 * device at will without giving the other program a notification of
 * the open or close.  Very useful with pose as it tends to open and
 * close the serial port frequently in the middle of a hotsync
 * operation.  It also removes the ordering which program had to be
 * run first and be waiting for input.
 *
 * Copyright (C) 2005 David Fries <david@fries.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; using version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *  GNU General Public License for more details. 
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * This program allows two programs on the same computer to talk
 * through a tty local null modem like cable.  Things like baud rates
 * are ignored, along with I'm sure other real hardware serial
 * interfaces, but the termios calls do not fail which allows at least
 * some programs to work without a problem using it.
 * 
 * The Unix Palm Pilot utilities use the serial port to connect to the
 * Palm Pilot's serial port for synchronization etc.  The Palm Pilot
 * emulator has the ability to use a Unix serial port as if it was the
 * serial port on the Palm Pilot.  The normal case as listed in the
 * emulator manual is to connect a physical serial null modem between
 * the two ports and talk over that.  It is closer to simulating the
 * connection to a real Palm handheld, but it seems silly to use up
 * two serial ports to talk from one program to another on the same
 * computer.  The Unix /dev/pty /dev/tty pairs provide a terminal
 * interface that allows the normal serial setup calls to be executed
 * ignoring what doesn't make sense for the real hardware such as baud
 * rate.  The problem is the order of operation.  The /dev/pty?? must
 * be opened first, then the /dev/tty?? and if either of them close
 * the sequence must be started over.  Unfortunately pose frequently
 * closes the serial port making it hard to complete even a single
 * hotsync operation.
 *
 * That is the reason for this program.  It opens two /dev/pty ports
 * and relays data from one to the other.  If a program closes the
 * /dev/tty port this program gets a read error which it retries after
 * a delay until the /dev/tty port is opened and data can again flow.
 * The other program need not be notified.  An alternate way to
 * delaying and retrying the read operation is to close the /dev/pty
 * and open it again.  This has the advantage that read will block
 * until there is data instead of returning EIO.  Unfortunately this
 * introduces a race condition, which with pose frequently opening and
 * closing the device comes up rather quickly.  The race condition is,
 * pose closes port, pty2pty gets EIO from read, pose opens port,
 * pty2pty closes port, pty2pty opens port.  The pty must be opened
 * first, and if the pty is closed while the tty is still open the pty
 * can't be opened again until the tty is closed.  At this point pose
 * has the tty open and pty2pty gets EIO each time it tries to open
 * the port.  This is a real race condition I'm seeing and polling
 * read is the only way I see to get around it.
 *
 * Future improvement might include some circular buffers.  For now it
 * is keep the program simple stupid, besides the kernel has buffers,
 * why not use them?
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/select.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

static const char *version = "$Id: pty2pty.cc,v 1.9 2005/07/04 16:44:50 david Exp $";

//#define DEBUG 1

int main(int argc, char **argv)
{
	char *device_name[2] = { NULL, NULL};
	int fd[2] = { -1, -1};
	const int buf_size = 4096;
	int buf[2][buf_size];
	int buf_count[2] = {0,0};
	int buf_loc[2] = {0,0};
	if(argc != 3)
	{
		fprintf(stderr, "Usage: %s /dev/pty?? /dev/pty??\n"
			"\tBoth pty devices are opened and data is forwarded"
			"from one to the other\n%s\n", argv[0], version);
		exit(1);
	}
	device_name[0] = argv[1];
	device_name[1] = argv[2];

	fd_set readfds, writefds;
	int events, result, i;

	/* Open the ports */
	for(i=0; i<2; ++i)
	{
		fd[i] = open(device_name[i], O_RDWR | O_NONBLOCK);
		if(fd[i] == -1)
		{
			fprintf(stderr, "Error opening port %s:%s\n",
				device_name[i], strerror(errno));
			exit(1);
		}
		struct termios t;
		if(tcgetattr(fd[i], &t)==-1)
		{
			perror("Error getting terminal settings\n");
			exit(1);
		}
		cfmakeraw(&t);
		if(tcsetattr(fd[i], TCSANOW, &t)==-1)
		{
			perror("Error setting terminal settings\n");
			exit(1);
		}
	}

	bool read_EIO[2] = {false,false};
	struct timeval timeout, *t_ptr;
	for(;;)
	{
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		t_ptr = NULL;
		/* Setup select. */
		for(i=0; i<2; ++i)
		{

			/* Setup select call. */
			/* Simple buffer, read or write, if there is data
			 * write, if there isn't data read, setup select
			 * operations for that direction only.
			 */
			if(read_EIO[i])
			{
				/* Do a timeout when the last pass returned
				 * and EIO error.
				 */
				t_ptr = &timeout;
				timeout.tv_sec = 0;
				/* Time timeout can't be too big or the
				 * pilot-xfer gets confused.  15 milliseconds
				 * was working pretty good.
				 */
				timeout.tv_usec = 15*1000;
			}
			else
			{
				if(buf_count[i])
					FD_SET(fd[(i+1)%2], &writefds);
				else
					FD_SET(fd[i], &readfds);
			}
		}
		events = select(1+(fd[0]>fd[1] ? fd[0] : fd[1]), &readfds,
			&writefds, NULL, t_ptr);
		if(events == -1)
		{
			perror("Error calling select");
			exit(1);
		}
		for(i=0; i<2; ++i)
		{
			/* Always clear the read error and try again next
			 * pass.
			 */
			read_EIO[i] = false;
			if(FD_ISSET(fd[i], &readfds))
			{
				result = read(fd[i], buf[i], buf_size);
				if(result == -1)
				{
					if(errno == EAGAIN)
					{
						perror(device_name[i]);
						break;
					}
					#if DEBUG
					fprintf(stderr, "Error reading %s: "
						"%s\n", device_name[i],
						strerror(errno));
					#endif
					read_EIO[i] = true;
					break;
				}
				else
				{
					buf_count[i] = result;
					buf_loc[i] = 0;
				}
			}
			if(FD_ISSET(fd[(i+1)%2], &writefds))
			{
				result = write(fd[(i+1)%2], buf[i]+buf_loc[i],
					buf_count[i]);
				if(result == -1)
				{
					#if DEBUG
					fprintf(stderr, "Error writing %s: "
						"%s\n", device_name[(i+1)%2],
						strerror(errno));
					#endif
					struct timespec req = {100*1000*1000,0};
					nanosleep(&req, NULL);
					break;
				}
				else
				{
					buf_loc[i] += result;
					buf_count[i] -= result;
				}
			}
		}
	}

	return 1;
}
